from __future__ import annotations

import dataclasses
import enum
from typing import TYPE_CHECKING, Final, final

if TYPE_CHECKING:
    from collections.abc import Mapping

    from ._state import State
    from ._symbol import Symbol

__all__: Final = ("Act", "Match", "Move", "Rules")


# A `Rules` object maps match rules to act rules.
type Rules = Mapping[Match, Act]


@final
@enum.unique
class Move(enum.Enum):
    """A movement command for a Turing machine."""

    LEFT = enum.auto()
    STAY = enum.auto()
    RIGHT = enum.auto()


@final
@dataclasses.dataclass(frozen=True, slots=True)
class Match:
    """
    A match rule.

    It matches when a Turing machine has read the given symbol, when
    the machine is in the given state.
    """

    symbol: Symbol
    state: State


@final
@dataclasses.dataclass(frozen=True, slots=True)
class Act:
    """
    An act rule.

    It commands a Turing machine to write the given symbol in the current cell,
    move the head as specified by `move`, and transition into the given state.
    """

    symbol: Symbol
    move: Move
    state: State



from __future__ import annotations

import collections
from typing import TYPE_CHECKING, Final, assert_never, final

from ._output import Output
from ._rules import Match, Move
from ._symbol import BLANK, Symbol

if TYPE_CHECKING:
    from ._input import Input
    from ._machine import Machine
    from ._tape import Tape

__all__: Final = ("run",)


@final
class _TapeHead:
    # An ever-extending tape-head.

    __slots__ = ("_index", "_tape")

    def __init__(self, tape: Tape, index: int, /) -> None:
        self._tape: Final = collections.deque(tape)

        # Initial "move" is arbitrary step-wise
        self._index = index
        for _ in range(self._index - len(self._tape) + 1):
            self._tape.append(BLANK)

    def export(self, /) -> tuple[Symbol, ...]:
        return tuple(self._tape)

    def read(self, /) -> Symbol:
        return self._tape[self._index]

    def write(self, symbol: Symbol, /) -> None:
        self._tape[self._index] = symbol

    def move(self, move: Move, /) -> None:
        match move:
            case Move.LEFT:
                if self._index < 1:
                    self._tape.appendleft(BLANK)
                else:
                    self._index -= 1
            case Move.RIGHT:
                self._index += 1
                if self._index == len(self._tape):
                    self._tape.append(BLANK)
            case Move.STAY:
                pass
            case never:
                assert_never(never)


@final
class _RunningMachine:
    # A "materialized" Turing machine.

    __slots__ = ("_state", "_tapehead", "halting_states", "rules")

    def __init__(self, machine: Machine, input_: Input, /) -> None:
        if frozenset(input_.tape) - frozenset(machine.alphabet):
            _msg = "Given tape contains symbols outside the alphabet"
            raise ValueError(_msg)

        self.rules: Final = machine.rules
        self.halting_states: Final = machine.halting_states
        self._tapehead: Final = _TapeHead(input_.tape, input_.initial_index)

        self._state = machine.initial_state

    def step(self, /) -> Output | None:
        current = Match(symbol=self._tapehead.read(), state=self._state)
        if current not in self.rules:
            _msg = f"No match for: {current}"
            raise RuntimeError(_msg)

        act = self.rules[current]
        self._tapehead.write(act.symbol)
        self._tapehead.move(act.move)
        self._state = act.state

        if self._state in self.halting_states:
            return Output(tape=self._tapehead.export(), state=self._state)
        return None


def run(*, machine: Machine, input_: Input) -> Output:
    """
    Run the given machine with the given input, and return the output.

    NOTE: If the machine wouldn't halt, neither would this function.
    """
    running_machine = _RunningMachine(machine, input_)
    while (output := running_machine.step()) is None:
        pass
    return output



from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING, Final, final

if TYPE_CHECKING:
    from ._tape import Tape

__all__: Final = ("Input",)


@final
@dataclasses.dataclass(kw_only=True, frozen=True, slots=True)
class Input:
    """
    A Turing machine input: the initial tape and an initial index.

    The initial index determines which cell of the tape to initially put the
    head on, counting from zero, left-to-right (default is zero i.e. the first
    cell).
    """

    tape: Tape
    initial_index: int = 0

    def __post_init__(self, /) -> None:
        if self.initial_index < 0:
            _msg = "Initial index must be a natural number"
            raise ValueError(_msg)



"""
Universal Turing Machine Emulator.

Abstract types:
- `Alphabet`
- `Tape`
- `Rules`

Elementary data types:
- `Symbol`
- `State`

Structured data types:
- `Match`
- `Act`
- `Input`
- `Output`
- `Machine`

Enumeration data types:
- `Move`

Pre-defined objects:
- `BLANK`

Functions:
- `run`
"""

from __future__ import annotations

from typing import Final

from ._alphabet import Alphabet
from ._input import Input
from ._machine import Machine
from ._output import Output
from ._rules import Act, Match, Move, Rules
from ._runner import run
from ._state import State
from ._symbol import BLANK, Symbol
from ._tape import Tape

__all__: Final = (
    "BLANK",
    "Act",
    "Alphabet",
    "Input",
    "Machine",
    "Match",
    "Move",
    "Output",
    "Rules",
    "State",
    "Symbol",
    "Tape",
    "run",
)

# Package metadata
__version__: Final = "1.0.0"



from __future__ import annotations

import dataclasses
from typing import Final, final

__all__: Final = ("State",)


@final
@dataclasses.dataclass(frozen=True, slots=True)
class State:
    """
    A Turing machine state.

    States have associated "code" values to separate qualitative identity
    (`==`) from numerical identity (`is)`, making it easier to copy/pickle
    states and objects composed of them.
    """

    code: str



from __future__ import annotations

import dataclasses
from typing import Final, final

__all__: Final = ("BLANK", "Symbol")


@final
@dataclasses.dataclass(frozen=True, slots=True)
class Symbol:
    """
    A symbol for a Turing machine.

    Symbols have associated one-letter "code" values for the following reasons:

    (1) To separate qualitative identity (`==`) from numerical identity (`is`),
    making it easier to copy/pickle symbols and objects composed of them.

    (2) To make it easy to convert tapes into strings and vice versa.
    """

    code: str

    def __post_init__(self, /) -> None:
        if len(self.code) != 1:
            _msg = "Symbols must have 1-character codes"
            raise ValueError(_msg)


# Represents the default symbol on an infinite tape.
BLANK: Final = Symbol(" ")



from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING, Final, final

if TYPE_CHECKING:
    from ._state import State
    from ._tape import Tape

__all__: Final = ("Output",)


@final
@dataclasses.dataclass(kw_only=True, frozen=True, slots=True)
class Output:
    """A Turing machine output: the final tape and the state it halted on."""

    tape: Tape
    state: State



from __future__ import annotations

import dataclasses
import warnings
from typing import TYPE_CHECKING, Final, final

from ._symbol import BLANK

if TYPE_CHECKING:
    from collections.abc import Collection

    from ._alphabet import Alphabet
    from ._rules import Rules
    from ._state import State

__all__: Final = ("Machine",)


@final
@dataclasses.dataclass(kw_only=True, frozen=True, slots=True)
class Machine:
    """A Turing machine description."""

    alphabet: Alphabet
    rules: Rules
    initial_state: State
    halting_states: Collection[State]

    def __post_init__(self, /) -> None:
        alphabet = frozenset((*self.alphabet, BLANK))
        halting_states = frozenset(self.halting_states)

        match_symbols = frozenset(match.symbol for match in self.rules)
        act_symbols = frozenset(act.symbol for act in self.rules.values())
        match_states = frozenset(match.state for match in self.rules)
        act_states = frozenset(act.state for act in self.rules.values())

        # Err if not all match rules' symbols are in the alphabet
        if match_symbols - alphabet:
            _msg = "Given match rules contain symbols outside the alphabet"
            raise ValueError(_msg)

        # Err if not all act rules' symbols are in the alphabet
        if act_symbols - alphabet:
            _msg = "Given act rules contain symbols outside the alphabet"
            raise ValueError(_msg)

        # Err if there is no match rule for the initial state
        if self.initial_state not in match_states:
            _msg = "No match rule matchs initial state"
            raise ValueError(_msg)

        # Err if there is not an act rule for each halting state
        if halting_states - act_states:
            _msg = "One or more halting states with no matching act rule"
            raise ValueError(_msg)

        # Warn if the alphabet contains anything besides BLANK
        if not self.alphabet:
            warnings.warn("Machine with blank-only alphabet", stacklevel=2)

        # Warn if a match rule matches a halting state
        if halting_states & match_states:
            warnings.warn(
                "One or more match rules matches a halting state", stacklevel=2
            )



from __future__ import annotations

from typing import TYPE_CHECKING, Final

if TYPE_CHECKING:
    from collections.abc import Collection

    from ._symbol import Symbol

__all__: Final = ("Alphabet",)


# Represents a finite set of symbols that a given Turing machine can recognize.
type Alphabet = Collection[Symbol]
# NOTE: `BLANK` is implicitly considered a member in any alphabet.



from __future__ import annotations

from typing import TYPE_CHECKING, Final

if TYPE_CHECKING:
    from collections.abc import Sequence

    from ._symbol import Symbol

__all__: Final = ("Tape",)


# Covers input and output tapes for Turing machines.
type Tape = Sequence[Symbol]
